import os
from typing import Any, ClassVar, Dict, Optional, Tuple, Union

from requests import Response, Session

from .resources import (
    AccessToken,
    Identity,
    Resource,
    UserValidationData,
    Verification,
)

API_URL = 'https://api.getmati.com'


class Client:

    base_url: ClassVar[str] = API_URL
    session: Session
    basic_auth_creds: Tuple[str, str]
    bearer_tokens: Dict[Union[None, str], AccessToken]

    # resources
    access_tokens: ClassVar = AccessToken
    identities: ClassVar = Identity
    user_validation_data: ClassVar = UserValidationData
    verifications: ClassVar = Verification

    def __init__(
        self, api_key: Optional[str] = None, secret_key: Optional[str] = None
    ):
        self.session = Session()
        api_key = api_key or os.environ['MATI_API_KEY']
        secret_key = secret_key or os.environ['MATI_SECRET_KEY']
        self.basic_auth_creds = (api_key, secret_key)
        self.bearer_tokens = {}
        Resource._client = self

    def get_valid_bearer_token(
        self, score: Optional[str] = None
    ) -> AccessToken:
        try:
            expired = self.bearer_tokens[score].expired
        except KeyError:
            expired = True
        if expired:  # renew token
            self.bearer_tokens[score] = self.access_tokens.create(score)
        return self.bearer_tokens[score]

    def get(self, endpoint: str, **kwargs: Any) -> Dict[str, Any]:
        return self.request('get', endpoint, **kwargs)

    def post(self, endpoint: str, **kwargs: Any) -> Dict[str, Any]:
        return self.request('post', endpoint, **kwargs)

    def request(
        self,
        method: str,
        endpoint: str,
        auth: Union[str, AccessToken, None] = None,
        token_score: Optional[str] = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        url = self.base_url + endpoint
        auth = auth or self.get_valid_bearer_token(token_score)
        headers = dict(Authorization=str(auth))
        response = self.session.request(method, url, headers=headers, **kwargs)
        self._check_response(response)
        return response.json()

    @staticmethod
    def _check_response(response: Response) -> None:
        if response.ok:
            return
        response.raise_for_status()



from base64 import b64encode


def basic_auth_str(username: str, password: str) -> str:
    auth = b64encode(f'{username}:{password}'.encode('utf-8'))
    return 'Basic ' + auth.decode('ascii')


def bearer_auth_str(access_token: str) -> str:
    return f'Bearer {access_token}'



import datetime as dt
from dataclasses import dataclass
from typing import ClassVar, Optional

from ..auth import basic_auth_str, bearer_auth_str
from .base import Resource


@dataclass
class AccessToken(Resource):
    """
    Based on: https://docs.getmati.com/#step-1-authentication
    """

    _endpoint: ClassVar[str] = '/oauth'

    token: str
    expires_at: dt.datetime
    score: Optional[str]
    user_id: Optional[str]

    @classmethod
    def create(cls, score: Optional[str] = None) -> 'AccessToken':
        data = dict(grant_type='client_credentials')
        endpoint = cls._endpoint
        if score:
            data['score'] = score
            endpoint += '/token'
        resp = cls._client.post(
            endpoint,
            data=data,
            auth=basic_auth_str(*cls._client.basic_auth_creds),
        )
        try:
            expires_in = resp['expiresIn']
        except KeyError:
            expires_in = resp['expires_in']
        expires_at = dt.datetime.now() + dt.timedelta(seconds=expires_in)
        try:
            user_id = resp['payload']['user']['_id']
        except KeyError:
            user_id = None
        return cls(
            user_id=user_id,
            token=resp['access_token'],
            expires_at=expires_at,
            score=score,
        )

    def __str__(self) -> str:
        return bearer_auth_str(self.token)

    @property
    def expired(self) -> bool:
        return self.expires_at < dt.datetime.now()



import datetime as dt
from dataclasses import dataclass, field
from typing import ClassVar, List, Optional, Union

from ..types import UserValidationFile
from .base import Resource
from .user_verification_data import UserValidationData


@dataclass
class Identity(Resource):
    """
    Based on: https://docs.getmati.com/#step-2-create-a-new-identity
    """

    _endpoint: ClassVar[str] = '/v2/identities'

    id: str
    dateCreated: dt.datetime
    dateUpdated: dt.datetime
    alive: Optional[bool]
    status: str
    annotatedStatus: Optional[str] = None
    user: Optional[str] = None
    metadata: Union[dict, List[str]] = field(default_factory=dict)
    fullName: Optional[str] = None
    facematchScore: Optional[float] = None

    @classmethod
    def create(cls, **metadata) -> 'Identity':
        resp = cls._client.post(cls._endpoint, json=dict(metadata=metadata))
        resp['id'] = resp.pop('_id')
        return cls(**resp)

    @classmethod
    def retrieve(cls, identity_id: str) -> 'Identity':
        endpoint = f'{cls._endpoint}/{identity_id}'
        resp = cls._client.get(endpoint)
        resp['id'] = resp.pop('_id')
        return cls(**resp)

    def refresh(self) -> None:
        identity = self.retrieve(self.id)
        for k, v in identity.__dict__.items():
            setattr(self, k, v)

    def upload_validation_data(
        self, user_validation_files: List[UserValidationFile]
    ) -> List[dict]:
        return UserValidationData.upload(self.id, user_validation_files)



import json
from typing import Any, BinaryIO, ClassVar, Dict, List, Tuple

from mati.types import UserValidationFile, ValidationInputType

from .base import Resource


def get_file_type(file: UserValidationFile) -> str:
    input_type = file.input_type
    if input_type == 'selfie-video':
        file_type = 'video'
    elif input_type == 'selfie-photo':
        file_type = 'selfie'
    else:
        file_type = 'document'
    return file_type


class UserValidationData(Resource):
    """
    Based on: https://docs.getmati.com/#step-3-upload-user-verification-data
    """

    _endpoint: ClassVar[str] = '/v2/identities/{identity_id}/send-input'

    @staticmethod
    def _append_file(
        files_metadata: List[Dict[str, Any]], file: UserValidationFile
    ):
        if file.input_type == ValidationInputType.document_photo:
            files_metadata.append(
                dict(
                    inputType=file.input_type,
                    group=file.group,
                    data=dict(
                        type=file.validation_type,
                        country=file.country,
                        page=file.page,
                        filename=file.filename,
                        region=file.region,
                    ),
                )
            )
        else:
            files_metadata.append(
                dict(
                    inputType=file.input_type,
                    data=dict(filename=file.filename),
                )
            )

    @classmethod
    def upload(
        cls, identity_id: str, user_validation_files: List[UserValidationFile]
    ) -> List[Dict[str, Any]]:
        endpoint = cls._endpoint.format(identity_id=identity_id)
        files_metadata: List[Dict[str, Any]] = []
        files_with_types: List[Tuple[str, BinaryIO]] = []
        for file in user_validation_files:
            cls._append_file(files_metadata, file)
            files_with_types.append((get_file_type(file), file.content))
        resp = cls._client.post(
            endpoint,
            data=dict(inputs=json.dumps(files_metadata)),
            files=files_with_types,
        )
        return resp



from dataclasses import dataclass, field
from enum import Enum
from typing import BinaryIO, Dict, List, Optional, Union


class SerializableEnum(str, Enum):
    def __str__(self):
        return self.value


class PageType(SerializableEnum):
    front = 'front'
    back = 'back'


class ValidationInputType(SerializableEnum):
    document_photo = 'document-photo'
    selfie_photo = 'selfie-photo'
    selfie_video = 'selfie-video'


class ValidationType(SerializableEnum):
    driving_license = 'driving-license'
    national_id = 'national-id'
    passport = 'passport'
    proof_of_residency = 'proof-of-residency'


@dataclass
class VerificationDocumentStep:
    id: str
    status: int
    error: Optional[str] = None
    data: Optional[Dict] = field(default_factory=dict)


@dataclass
class VerificationDocument:
    country: str
    region: str
    photos: List[str]
    steps: List[VerificationDocumentStep]
    type: str
    fields: Optional[dict] = None


@dataclass
class UserValidationFile:
    filename: str
    content: BinaryIO
    input_type: Union[str, ValidationInputType]
    validation_type: Union[str, ValidationType] = ''
    country: str = ''  # alpha-2 code: https://www.iban.com/country-codes
    region: str = ''  # 2-digit US State code (if applicable)
    group: int = 0
    page: Union[str, PageType] = PageType.front



from typing import ClassVar, Optional

import iso8601


class Resource:
    _client: ClassVar['mati.Client']  # type: ignore
    _endpoint: ClassVar[str]
    _token_score: ClassVar[Optional[str]] = None

    def __post_init__(self) -> None:
        for attr, value in self.__dict__.items():
            if attr.startswith('date'):
                setattr(self, attr, iso8601.parse_date(value))



__all__ = [
    'AccessToken',
    'Identity',
    'Resource',
    'UserValidationData',
    'Verification',
]

from .access_tokens import AccessToken
from .base import Resource
from .identities import Identity
from .user_verification_data import UserValidationData
from .verifications import Verification



from dataclasses import dataclass, field
from typing import Any, ClassVar, Dict, List, Optional

from ..types import VerificationDocument, VerificationDocumentStep
from .base import Resource


@dataclass
class Verification(Resource):
    _endpoint: ClassVar[str] = '/api/v1/verifications'
    _token_score: ClassVar[str] = 'identity'

    id: str
    expired: bool
    steps: list
    documents: List[VerificationDocument]
    metadata: Dict[str, Dict[str, str]]
    identity: Dict[str, str] = field(default_factory=dict)
    hasProblem: Optional[bool] = None
    computed: Optional[Dict[str, Any]] = None

    @classmethod
    def retrieve(cls, verification_id: str) -> 'Verification':
        endpoint = f'{cls._endpoint}/{verification_id}'
        resp = cls._client.get(endpoint, token_score=cls._token_score)
        docs = []
        for doc in resp['documents']:
            doc['steps'] = [
                VerificationDocumentStep(**step) for step in doc['steps']
            ]
            docs.append(VerificationDocument(**doc))
        resp['documents'] = docs
        return cls(**resp)
